import pdfriend.classes.wrappers as wrappers
import pypdf
import getpass

def decrypt(infile: str, outfile: str):
    pdf = pypdf.PdfReader(infile)

    if pdf.is_encrypted:
        password = getpass.getpass("password: ")

        pdf.decrypt(password)

    wrapper = wrappers.PDFWrapper(pages = pdf.pages)
    wrapper.write(outfile)




from pdfriend.classes.platforms import Platform
import shutil

def cache(subcommand: str):
    if subcommand == "clear":
        if Platform.BackupDir.exists():
            shutil.rmtree(Platform.BackupDir.as_posix())

        Platform.BackupDir.mkdir()



import argparse
from pdfriend.classes.config import Config
from pdfriend.classes.platforms import Platform
import pdfriend.classes.cmdparsers as cmdparsers
import pdfriend.classes.exceptions as exceptions
import pdfriend.classes.info as info
import pdfriend.commands as commands
import pdfriend.utils as utils


program_info = info.ProgramInfo(
    info.CommandInfo("version", "v", descr = """ | -v | --version
    prints the current version of pdfriend
    """),
    info.CommandInfo("help", "h", descr = """ [command?]
        display help message. If given a command, it will only display the help message for that command.

        examples:
            pdfriend help merge
                prints out the help blurb for the merge command
            pdfriend help
                prints out generic help message, with a command list
    """),
    info.CommandInfo("merge", "m", descr = """ [filename1] [filename2?] ... [-o|--outfile outfile?=pdfriend_output.pdf] [-q|--quality quality?=100]
        merge the given files into one pdf. It can handle multiple pdfs, as well convert and merge png and jpg images. Glob patterns are also supported. You can specify the output filename using the -o or --outfile flag, otherwise it defaults to pdfriend_output.pdf. You can also specify the quality when images are converted to pdfs via the -q or --quality flag. It's an integer going from 0 to 100, 100 is no lossy compression and 0 is full lossy compression.

    examples:
        pdfriend merge pdf1.pdf img.png pdf2.pdf -o merged.pdf
            merges all of those into merged.pdf, preserving the quality of img.png
        pdfriend merge folder_name/* -o merged.pdf -q 50
            merges all files in directory folder_name into merged.pdf and compresses the images by 50%.
        pdfriend merge pdf1.pdf folder_name/* img.jpg pdf2.pdf -o apricot.pdf
            merges every file given, including all files in folder_name, into apricot.pdf
    """),
    info.CommandInfo("split", "s", descr = """ [filename] [pages] ... [-o|--outfile outfile?=pdfriend_output]
        split the given file at the given points. Every point is included in the part after, not before it.

        examples:
            pdfriend split in.pdf 5 -o parts
                splits in.pdf into one part with pages 1-4 and another with 5-end and puts them in a directory named parts
            pdfriend split input.pdf 4,7
                splits input.pdf into 3 parts, one has pages 1-3, another 4-6 and another 7-end and puts them in a directory named pdfriend_output
            pdfriend split thing.pdf 8:11
                splits thing.pdf into 4 parts, one has pages 1-7, another has page 8, the other page 9, the other page 10, and the other pages 11-end and puts them in a directory named pdfriend_output
            pdfriend split infile.pdf all -o pages
                splits infile.pdf into individual pages and puts them in a directory named pages
    examples:
        pdfriend merge pdf1.pdf img.png pdf2.pdf -o merged.pdf
            merges all of those into merged.pdf, preserving the quality of img.png
        pdfriend merge folder_name/* -o merged.pdf -q 50
            merges all files in directory folder_name into merged.pdf and compresses the images by 50%.
        pdfriend merge pdf1.pdf folder_name/* img.jpg pdf2.pdf -o apricot.pdf
            merges every file given, including all files in folder_name, into apricot.pdf
    """),
    info.CommandInfo("edit", "e", descr = """ [filename] [-I|--import command_file?]
        edit the selected file in place, using a set of subcommands. After launching the edit shell, you can type h or help to list the subcommands. You can import pdfriend edit subcommands from text files using -I. The text files must have the edit subcommands as you would write them on the shell, separated by newlines. Try running some commands in the edit shell and then exporting them using the x command.

        examples:
            pdfriend edit notes.pdf
                launches the edit shell on notes.pdf
            pdfriend edit fried.pdf -I some_commands.txt
                executes the commands in some_commands.txt and immediately exits the edit shell
    """),
    info.CommandInfo("tinker", "t", descr = """ [filename] [-I|--import command_file?]
        edit the selected file page by page. Currently there is only the ability to show and export images

        examples:
            pdfriend tinker notes.pdf
                launches the tinker shell on notes.pdf
            pdfriend tinker fried.pdf -I some_commands.txt
                executes the commands in some_commands.txt and immediately exits the tinker shell
    """),
    info.CommandInfo("invert", "i", descr = """ [filename] [-o|--outfile outfile?=pdfriend_output.pdf] [-i|--inplace?]
        create a PDF file with the pages of the input file, but in inverted order. Adding -i or --inplace will make it so the input file is modified, instead of creating a new one.

        examples:
            pdfriend invert puppy.pdf -o puppy-inv.pdf
                inverts the pages of puppy.pdf and saves to puppy-inv.pdf
            pdfriend invert kitty.pdf -i
                inverts the pages of kitty.pdf
    """),
    info.CommandInfo("cache", "c", descr = """ [subcommand]
        for managing the pdfriend cache. Currently, you can only clear it

        examples:
            pdfriend cache clear
                clears the cache
    """),
    info.CommandInfo("weave", "w", descr = """ [filename_0] [filename_1] [-o|--outfile?=pdfriend_output.pdf]
        combines two PDFs such that the first fills the odd pages and the second the even pages of the output.

        examples:
            pdfriend weave inf0.pdf inf1.pdf
                weaves the two PDFs and saves the output to pdfriend_output.pdf
            pdfriend weave k.pdf l.pdf -o weaved.pdf
                weaves the two PDFs and saves the output to weaved.pdf
    """),
    info.CommandInfo("get", "g", descr = """ [filename] [pages] [-o|--outfile?=pdfriend_output.pdf] [-i|--inplace?]
        takes the given pages out of the PDF and puts them in the specified output file.

        examples:
            pdfriend get comic.pdf 34 -o comic-34.pdf
                gets the 34th page of comic.pdf and puts it into comic-34.pdf
            pdfriend get sth.pdf 1,2,7
                gets the pages 1,2,7 from sth.pdf and puts them into pdfriend_output.pdf
            pdfriend get reef.pdf 3:54 -o reef_main.pdf
                gets pages 3 through 54 (INCLUDING 54) and puts them into reef_main.pdf
            pdfriend get shrimp_manual.pdf 5: -o shrimp_manual_trimmed.pdf
                gets pages from 5 to the end of shrimp_manual.pdf and puts them into shrimp_manual_trimmed.pdf
            pdfriend get notes.pdf 3,4:15,18,20 -i
                gets pages 3,4,5,...,15,18,20 from notes.pdf and MODIFIES the file such that it only has those pages
    """),
    info.CommandInfo("encrypt", "n", descr = """ [filename] [-o|--outfile?=pdfriend_output.pdf] [-i|--inplace?]
        creates an encrypted PDF file using a provided password. Adding -i or --inplace will make it so that the file itself is encrypted.

        examples:
            pdfriend encrypt not-sus.pdf -i
                encrypts not-sus.pdf in-place. Make sure you remember the password, as it will be overwritten!
            pdfriend encrypt balance.pdf -o balance-encrypted.pdf
                encrypts balance.pdf and saves to balance-encrypted.pdf.
            pdfriend encrypt acct.pdf
                encrypts acct.pdf and saves to pdfriend_output.pdf.
    """),
    info.CommandInfo("decrypt", "d", descr = """ [filename] [-o|--outfile?=pdfriend_output.pdf] [-i|--inplace?]
        decrypts an encrypted PDF file using a provided password. Adding -i or --inplace will make it so that the file itself is decrypted. If the file is not encrypted, it will just be copied.

        examples:
            pdfriend decrypt not-sus.pdf -i
                decrypts not-sus.pdf in-place.
            pdfriend decrypt balance.pdf -o balance-decrypted.pdf
                decrypts balance.pdf and saves to balance-decrypted.pdf.
            pdfriend decrypt acct.pdf
                decrypts acct.pdf and saves to pdfriend_output.pdf.
    """),
    info.CommandInfo("metadata", "M", descr = """ [filename] [--get key?] [--set key_val_pairs?] [--pop keys?]
        manages PDF metadata. Using no extra flags, it will print the key-value pairs. You can use --get to print the value of a specific key and --set to set values for keys, or --pop to delete them.

        examples:
            pdfriend metadata some.pdf
                prints the metadata of some.pdf.
            pdfriend metadata thing.pdf --get /Author
                prints the name of the author of the PDF, if that field has been set.
            pdfriend metadata stolen.pdf --set /Author=me
                sets the author of stolen.pdf to "me". BEWARE: This will overwrite the PDF, unlike most of the other pdfriend commands.
            pdfriend metadata cnp.pdf --set  "/Title=Crime And Punishment"
                sets the title to that. Note that you need the quotes here, else your shell will interpret the words as different arguments.
            pdfriend metadata phys1.pdf --set "/Title=University Physics with Modern Physics,/Author=H. Young and R. Freedman"
                sets the author of phys1.pdf to "H. Young and R. Freedman" and its title to you-know-what.
            pdfriend metadata embarassing_fanfic.pdf --pop /Author
                removes (!) the PDF's author field.
            pdfriend metadata mystery.pdf --pop /Author,/Producer
                removes the PDF's author and producer fields.
    """),
    foreword = """pdfriend: a command line utility for easily modifying PDF files
    usage: pdfriend [command] [arguments?] [-d|--debug?] (note that options in [] are required and options in [?] are not). Use -d or --debug for more detailed error messages.
    the following commands are available:
    """,
    postword = """use pdfriend help [command] to get the instructions for particular commands"""
)

def run_pdfriend(args):
    try:
        cmd_parser = cmdparsers.CmdParser.FromArgs(
            program_info,
            args.commands,
            no_command_message = "No command specified! Use pdfriend help to get a list of the available commands"
        )

        short = cmd_parser.short()

        if short == "v" or args.version:
            print(commands.version())
        elif short == "h" or args.help:
            command_to_display = cmd_parser.next_str_or(None)
            print(program_info.help(command_to_display))
        elif short == "m":
            if len(cmd_parser.args) == 0:
                print("You need to specify at least one file or pattern to be merged")
                return

            commands.merge(cmd_parser.args, args.outfile, args.quality)
        elif short == "e":
            pdf = cmd_parser.next_pdf("filename")

            # I have to do this because args.import is a syntax error
            # for some reason. Must be messing with the import keyword?
            import_file = args.__getattribute__("import")
            import_args = None
            if import_file is not None:
                import_args = cmdparsers.to_shell_import(import_file)

            commands.edit(pdf, commands = import_args, open_pdf = Config.OpenPDFs)
        elif short == "t":
            pdf = cmd_parser.next_pdf("filename")

            # I have to do this because args.import is a syntax error
            # for some reason. Must be messing with the import keyword?
            import_file = args.__getattribute__("import")
            import_args = None
            if import_file is not None:
                import_args = cmdparsers.to_shell_import(import_file)

            commands.tinker(pdf, commands = import_args, open_pdf = Config.OpenPDFs)
        elif short == "i":
            pdf = cmd_parser.next_pdf("filename")

            if args.inplace:
                args.outfile = pdf.source

            commands.invert(pdf, args.outfile)
        elif short == "c":
            subcommand = cmd_parser.next_str("subcommand")

            commands.cache(subcommand)
        elif short == "w":
            pdf_0 = cmd_parser.next_pdf("filename_0")
            pdf_1 = cmd_parser.next_pdf("filename_1")

            commands.weave(pdf_0, pdf_1, args.outfile)
        elif short == "s":
            pdf = cmd_parser.next_pdf("filename")
            slice = cmd_parser.next_pdf_slice(pdf, name = "pages")

            commands.split(pdf, slice, args.outfile)
        elif short == "g":
            pdf = cmd_parser.next_pdf("filename")
            slice = cmd_parser.next_pdf_slice(pdf, name = "pages")

            if args.inplace:
                args.outfile = pdf.source

            commands.get(pdf, slice, args.outfile)
        elif short == "n":
            pdf = cmd_parser.next_pdf("filename")

            if args.inplace:
                args.outfile = pdf.source

            commands.encrypt(pdf, args.outfile)
        elif short == "d":
            infile = cmd_parser.next_file("filename")

            if args.inplace:
                args.outfile = infile

            commands.decrypt(infile, args.outfile)
        elif short == "meta":
            pdf = cmd_parser.next_pdf("filename")

            commands.metadata(pdf, args.get, args.set, args.pop)
        else:
            print(f"command \"{command}\" not recognized")
            print("use pdfriend help for a list of the available commands")
    except exceptions.ExpectedError as e:
        print(e)
    except Exception as e:
        utils.print_unexpected_exception(e, Config.Debug)


def main():
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument("commands", type=str, nargs="*")

    parser.add_argument("-h", "--help", action="store_true")
    parser.add_argument("-v", "--version", action="store_true")
    parser.add_argument("-d", "--debug", action="store_true")

    parser.add_argument("-o", "--outfile", type=str, default="pdfriend_output")
    parser.add_argument("-i", "--inplace", action="store_true")
    parser.add_argument("-q", "--quality", type=int, default=100)
    parser.add_argument("-I", "--import", type=str, default=None)

    parser.add_argument("--get", type=str, default=None)
    parser.add_argument("--set", type=str, default=None)
    parser.add_argument("--pop", type=str, default=None)

    args = parser.parse_args()

    Platform.Init()
    Config.Debug = args.debug

    run_pdfriend(args)




import platformdirs
import pathlib
import shutil
import subprocess
import platform
import os

def ensuredir(path: pathlib.Path):
    if not path.exists():
        path.mkdir()


class Platform:
    ConfigDir: pathlib.Path = platformdirs.user_config_path().joinpath("pdfriend")
    CacheDir: pathlib.Path = platformdirs.user_cache_path().joinpath("pdfriend")
    TempDir: pathlib.Path = CacheDir.joinpath("temp")
    BackupDir: pathlib.Path = CacheDir.joinpath("backups")

    @classmethod
    def Init(cls): # make sure the system directories exist
        ensuredir(cls.ConfigDir)
        ensuredir(cls.CacheDir)
        ensuredir(cls.BackupDir)

        if cls.TempDir.exists(): # temp dir always cleared on startup
            shutil.rmtree(cls.TempDir.as_posix())
        cls.TempDir.mkdir()

    @classmethod
    def NewTemp(cls, path_name: str) -> pathlib.Path:
        return cls.TempDir.joinpath(path_name)

    @classmethod
    def NewBackup(cls, path_name: str) -> pathlib.Path:
        return cls.BackupDir.joinpath(path_name)

    @classmethod
    def OpenFile(self, filename: str):
        sys = platform.system()
        if sys == "Linux":
            subprocess.Popen(["xdg-open", filename])
        elif sys == "Darwin":
            subprocess.Popen(["open", filename])
        elif sys == "Windows":
            os.startfile(filename)




import pdfriend.classes.wrappers as wrappers
import pdfriend.classes.exceptions as exceptions
import pdfriend.classes.cmdparsers as cmdparsers
import pdfriend.classes.shells as shells
import pdfriend.classes.info as info
from pdfriend.classes.platforms import Platform
from pdfriend.classes.config import Config
import pdfriend.utils as utils
import pathlib


program_info = info.ProgramInfo(
    info.CommandInfo("help", "h", descr = """[command?]
    display help message. If given a command, it will only display the help message for that command.

    examples:
        help rotate
            displays the help blurb for the rotate command
        help exit
            displays the help blurb for the exit command
    """),
    info.CommandInfo("exit", "e", descr = """
    exits the edit mode
    """),
    info.CommandInfo("rotate", "r", descr = """[page_numbers] [angle]
    rotates page clockwise with the given numbers (starting from 1) by the given angle (in degrees). Can use negative angles to rotate counter-clockwise. DO NOT put extra spaces between the page numbers!

    examples:
        r 34 1.2
            rotates page 34 clockwise by 1.2 degrees
        r 1,3,8 -4
            rotates pages 1,3 and 8 counter-clockwise by 4 degrees
        r 3:18 90
            rotates pages 3 through 18 (INCLUDING 18) clockwise by 90 degrees
        r 1,13,5:7,2 54
            rotates pages 1,2,5,6,7,13 clockwise by 54 degrees
        r all -90
            rotates all pages counter-clockwise by 90 degrees
    """),
    info.CommandInfo("delete", "d", descr = """[page_numbers]
    deletes all specified pages. DO NOT put extra spaces between the page numbers!

    examples:
        d 7
            deletes page 7
        d 4,8,1
            deletes pages 1, 4 and 8
        d 6:66
            deletes pages 6 through 66 (INCLUDING 66)
        d :13
            deletes all pages up to and including 13
        d 4,17,3:6
            deletes pages 3,4,5,6 and 17
    """),
    info.CommandInfo("swap", "s", descr = """[page_0] [page_1]
    swaps page_0 and page_1.
    """),
    info.CommandInfo("move", "m", descr = """[source] [destination]
    move source to BEFORE destination, taking its place.

    examples:
        m 3 17
            moves page 3 to page 17, pushing back the pages from 17 onward
        m 83 1
            moves page 83 to the beginning of the document
    """),
    info.CommandInfo("push", "p", descr = """[pages] [offset]
    pushes the specified pages by offset pages (offset can be negative).

    examples:
        p 3 7
            moves page 3 to 7 pages down, i.e. to page 10.
        p 4,9,2 1
            moves pages 2,4,9 by 1 page.
        p 5:8 -3
            moves pages 5,6,7,8 to 3 pages BACK.
        p 5,6,90:94 5
            moves pages 5,6,90,91,92,93,94 to be 5 pages down.
        p :5 4
            moves pages 1,2,3,4,5 to be 4 pages down.
        p 67: -7
            move pages from 67 to the end of the PDF to be 7 pages back.
        p 70: 5
            FAILS. 70: includes the end of the PDF, and you can't move that further down.
    """),
    info.CommandInfo("undo", "u", descr = """[number?]
    undo the previous [number] commands.

    examples:
        u
            undoes the previous command
        u 3
            undoes the previous 3 commands
        u all
            undoes all commands issued this session (reverts document fully)
    """),
    info.CommandInfo("export", "x", descr = """[filename?=pdfriend_edit.txt]
    exports all the commands you ran into a text file

        examples:
            x
                exports your commands to pdfriend_edit.txt
            x out.txt
                exports your commands to out.txt
    """),
    foreword = "pdfriend edit shell for quick changes. Commands:",
    postword = "use h [command] to learn more about a specific command"
)


class EditRunner(shells.ShellRunner):
    def __init__(self, pdf: wrappers.PDFWrapper, open_pdf: bool = True):
        backup_path = pdf.backup()
        print(f"editing {pdf.source}\nbackup created in {backup_path}")

        self.pdf = pdf
        self.backup_path = backup_path

        if open_pdf:
            Platform.OpenFile(pdf.source)

    def parse(self, arg_str) -> list[str]:
        return utils.parse_command_string(arg_str)

    def run(self, args: list[str]):
        cmd_parser = cmdparsers.CmdParser.FromArgs(
            program_info,
            args,
            no_command_message = "No command specified! Type h or help for a list of the available commands"
        )
        short = cmd_parser.short()

        if short == "h":
            subcommand = cmd_parser.next_str_or(None)
            print(program_info.help(subcommand))

            # this is to prevent rewriting the file and appending
            # the command to the command stack
            raise exceptions.ShellContinue()
        elif short == "e":
            raise exceptions.ShellExit()
        elif short == "r":
            pages = cmd_parser.next_pdf_slice(self.pdf, "pages")
            angle = cmd_parser.next_float("angle")

            if len(pages) == 0:
                return
            # the slice is sorted, so if any pages are out of range, it'll
            # either be the first or the last one, probably the last
            self.pdf.raise_if_out_of_range(pages[-1])
            self.pdf.raise_if_out_of_range(pages[0])

            for page in pages:
                self.pdf.rotate_page(page, angle)
        elif short == "d":
            pages = cmd_parser.next_pdf_slice(self.pdf, "pages")

            for page in pages:
                self.pdf.pop_page(page)
        elif short == "s":
            page_0 = cmd_parser.next_int("page_0")
            self.pdf.raise_if_out_of_range(page_0)
            page_1 = cmd_parser.next_int("page_1")
            self.pdf.raise_if_out_of_range(page_1)

            self.pdf.swap_pages(page_0, page_1)
        elif short == "m":
            source = cmd_parser.next_int("source")
            self.pdf.raise_if_out_of_range(source)
            destination = cmd_parser.next_int("destination")
            self.pdf.raise_if_out_of_range(destination)

            page = self.pdf.pages.pop(source - 1)
            self.pdf.pages.insert(destination - 1, page)
        elif short == "p":
            pages = cmd_parser.next_pdf_slice(self.pdf, "pages")
            offset = cmd_parser.next_int("offset")

            last_page_before = pages[-1]
            last_page_after = last_page_before + offset

            if last_page_after > self.pdf.len(): # only check last page, as the slice is sorted
                raise exceptions.ExpectedError(
                    f"can't move page {last_page_before} to {last_page_after}, as it's outside the PDF (number of pages: {self.pdf.len()})"
                )

            if offset > 0:
                pages = pages[::-1]

            for page in pages:
                p = self.pdf.pages.pop(page - 1)
                self.pdf.pages.insert(page + offset - 1, p)
        elif short == "u":
            # arg will be converted to int, unless it's "all". Defaults to 1
            num_of_commands = cmd_parser.next_typed_or(
                "int or \"all\"", lambda s: s if s == "all" else int(s),
                1 # default value
            )

            raise exceptions.ShellUndo(num_of_commands)
        elif short == "x":
            filename = cmd_parser.next_str_or("pdfriend_edit.txt")

            raise exceptions.ShellExport(filename)

    def reset(self):
        self.pdf.reread(self.backup_path)

    def save(self):
        self.pdf.write()

    def exit(self):
        pass


def edit(
    pdf: wrappers.PDFWrapper,
    commands: list[str] | None = None,
    open_pdf: bool = True
):
    edit_shell = shells.Shell(
        runner = EditRunner(pdf, open_pdf = open_pdf)
    )

    edit_shell.run(commands)




import pypdf
import datetime
import pathlib
import shutil
import pdfriend.classes.exceptions as exceptions
from typing import Self
from PIL import Image
from pdfriend.classes.platforms import Platform

class PDFWrapper:
    def __init__(self,
        source: pathlib.Path = None,
        pages: list[pypdf.PageObject] = None,
        metadata: pypdf.DocumentInformation = None,
        reader: pypdf.PdfReader = None,
    ):
        self.source = source
        self.pages = pages or []
        self.metadata = metadata
        self.reader = reader
        if metadata is not None:
            self.metadata = dict(metadata)

    def __getitem__(self, num: int) -> pypdf.PageObject:
        return self.pages[num - 1]

    def __setitem__(self, num: int, page: pypdf.PageObject):
        self.pages[num - 1] = page

    @classmethod
    def Read(cls, filename: str):
        pdf = pypdf.PdfReader(filename)

        return PDFWrapper(
            source = pathlib.Path(filename),
            pages = list(pdf.pages),
            metadata = pdf.metadata,
            reader = pdf,
        )

    def reread(self, source: pathlib.Path, keep_metadata: bool = True):
        new_pdf = PDFWrapper.Read(source)
        self.pages = new_pdf.pages
        if not keep_metadata:
            self.metadata = new_pdf.metadata

        return self

    def len(self):
        return len(self.pages)

    def slice(self, slice_str: str) -> list[int]:
        if slice_str == "all":
            return list(range(1, self.len()))

        result = []
        for subslice in slice_str.split(","):
            if ":" not in subslice:
                page_num = int(subslice)
                if page_num < 1 or page_num > self.len():
                    continue

                result.append(page_num)
                continue

            split_subslice = subslice.split(":")
            first, last = split_subslice[0], split_subslice[-1]

            # such that n: means n:end and :n means 1:n
            first = 1 if first == "" else int(first)
            last = self.len() if last == "" else int(last)

            # making sure the subrange is within bounds
            lower = max(min(first, last), 1)
            upper = min(max(first, last), self.len())

            result.extend(list(range(lower, upper + 1)))

        return sorted(list(set(result)))

    def subset(self, slice: str | list[str]) -> Self:
        if isinstance(slice, str):
            slice = self.slice(slice)

        return PDFWrapper(
            source = self.source,
            pages = [self.pages[idx - 1] for idx in slice],
            metadata = self.metadata
        )

    def raise_if_out_of_range(self, page_num: int):
        if page_num >= 1 and page_num <= self.len():
            return
        raise exceptions.ExpectedError(
            f"page {page_num} doesn't exist in the PDF (total pages: {self.len()})"
        )

    def rotate_page(self, page_num: int, angle: float) -> Self:
        int_angle = int(angle)
        if int_angle % 90 == 0:
            self[page_num].rotate(int_angle)
            return self

        rotation = pypdf.Transformation().rotate(angle)
        self[page_num].add_transformation(rotation)
        return self

    def pop_page(self, page_num: int) -> pypdf.PageObject:
        return self.pages.pop(page_num - 1)

    def append_page(self, page: pypdf.PageObject) -> Self:
        self.pages.append(page)
        return self

    def extend(self, pages: list[pypdf.PageObject]) -> Self:
        self.pages.extend(pages)
        return self

    def swap_pages(self, page_num_0: int, page_num_1: int) -> Self:
        temp = self[page_num_0]
        self[page_num_0] = self[page_num_1]
        self[page_num_1] = temp

        return self

    def merge_with(self, other: Self) -> Self:
        self.pages.extend(other.pages)

        return self

    def invert(self) -> Self:
        self.pages = self.pages[::-1]

        return self

    def weave_with(self, other: Self) -> Self:
        result = PDFWrapper()
        for odd_page, even_page in zip(self.pages, other.pages):
            result.extend([odd_page, even_page])

        if self.len() > other.len():
            result.extend(self.pages[other.len():])
        elif other.len() > self.len():
            result.extend(other.pages[self.len():])

        return result

    def to_writer(self):
        writer = pypdf.PdfWriter()
        if self.reader is not None and False: # FIXME
            writer.clone_document_from_reader(self.reader)
        else:
            for page in self.pages:
                writer.add_page(page)

        return writer

    def write(self, filename: str = None, keep_metadata = True):
        if filename is None:
            filename = self.source

        writer = self.to_writer()
        if keep_metadata and self.metadata is not None:
            writer.add_metadata(self.metadata)

        writer.write(
            pathlib.Path(filename).with_suffix(".pdf")
        )

    def backup(self, name: str | pathlib.Path = None, copy: bool = True) -> pathlib.Path:
        if name is None:
            name = self.source

        if not isinstance(name, pathlib.Path):
            name = pathlib.Path(name)

        now: str = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M")
        backup_file: pathlib.Path = Platform.NewBackup(
            f"{name.stem}_{now}.pdf"
        )

        # prefer to just copy the file from the source if possible
        if copy and self.source.is_file():
            shutil.copyfile(self.source, backup_file)
        else:
            self.write(backup_file)

        return backup_file


def convert_to_rgb(img_rgba: Image.Image):
    try:
        img_rgba.load()
        _, _, _, alpha = img_rgba.split()

        img_rgb = Image.new("RGB", img_rgba.size, (255, 255, 255))
        img_rgb.paste(img_rgba, mask=alpha)

        return img_rgb
    except (IndexError, ValueError):
        return img_rgba


class ImageWrapper:
    def __init__(self, images: list[Image.Image]):
        self.images = [convert_to_rgb(image) for image in images]

    @classmethod
    def FromFiles(cls, filenames: list[str]) -> Self:
        return ImageWrapper([Image.open(filename) for filename in filenames])

    def equalize_widths(self):
        max_width = max([image.size[0] for image in self.images])

        for i, image in enumerate(self.images):
            width, height = image.size

            scale = max_width / width

            self.images[i] = image.resize((max_width, int(height * scale)))

    def write(self, outfile: str, quality: int | float):
        self.images[0].save(
            outfile,
            "PDF",
            optimize=True,
            quality=int(quality),
            save_all=True,
            append_images=self.images[1:],
        )



# I do not use sets or dictionaries here because they're not necessary
# the number of commands is not expected to be large

class CommandInfo:
    def __init__(self, primary_name: str, short_name: str, alt_names: set[str] = None, descr: str = ""):
        self.primary_name = primary_name
        self.short_name = short_name
        self.alt_names = [short_name] + (alt_names or [])
        self.all_names = [primary_name] + self.alt_names
        self.descr = descr

    def help(self) -> str:
        return f"{' | '.join(self.all_names)} {self.descr}"


class ProgramInfo:
    def __init__(self, *commands: CommandInfo, foreword: str = "", postword: str = ""):
        self.foreword = foreword
        self.commands = commands
        self.postword = postword

    def get_command_info(self, command_name: str) -> CommandInfo | None:
        return next(
            (
                command for command in self.commands
                if command_name in command.all_names
            ),
            None
        )

    def help(self, command_name: str | None) -> str:
        if command_name is not None:
            command_info = self.get_command_info(command_name)
            if command_info is None:
                return None

            return command_info.help()

        column_names = ["command", "alternatively"]
        spaces = 2
        spacing = " " * spaces

        width = max(
            [len(column_names[0])] + [
                len(command.primary_name) for command in self.commands
            ]
        ) + spaces

        column_heads = f"{column_names[0].rjust(width)}{spacing}{column_names[1]}"
        command_list = "\n".join([
            f"{command.primary_name.rjust(width)}{spacing}{','.join(command.alt_names)}"
            for command in self.commands
        ])

        return "\n".join([
            self.foreword,
            column_heads,
            command_list,
            self.postword,
        ])




class ShellExit(Exception):
    pass

class ShellContinue(Exception):
    pass

class ShellUndo(Exception):
    def __init__(self, num: str | int):
        self.num = num
        super().__init__()

class ShellExport(Exception):
    def __init__(self, filename: str | int):
        self.filename = filename
        super().__init__()

class ExpectedError(Exception):
    pass



import re
import traceback

def print_unexpected_exception(exception: Exception, debug: bool = False):
    print("unexpected error occured:")
    if debug:
        traceback.print_exception(exception)
    else:
        print(exception)

whitespace_pattern = re.compile(r"\s+")

def parse_command_string(command_string: str) -> list[str]:
    return re.split(whitespace_pattern, command_string)




import pdfriend.classes.wrappers as wrappers
import pdfriend.classes.exceptions as exceptions
import pdfriend.utils as utils
from pdfriend.classes.config import Config
from abc import ABC, abstractmethod

def input_generator():
    while True:
        yield input("")


class ShellRunner(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def parse(self, arg_str: str) -> list[str]:
        pass

    @abstractmethod
    def run(self, args: list[str]):
        pass

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def save(self):
        pass

    @abstractmethod
    def exit(self):
        pass


class Shell:
    def __init__(self, runner: ShellRunner):
        self.command_stack = []
        self.runner = runner

    def run_command(self, args: list[str]):
        args = self.runner.parse(args)
        self.runner.run(args)
        self.command_stack.append(args)
        self.runner.save()

    def undo(self, num_of_commands: int | str):
        if num_of_commands == "all":
            self.command_stack = []
        else:
            self.command_stack = self.command_stack[:-num_of_commands]

        self.runner.reset()
        for args in self.command_stack:
            self.runner.run(args)

        self.runner.save()

    def export(self, filename: str):
        with open(filename, "w") as outfile:
            outfile.write("\n".join([
                " ".join(args) for args in self.command_stack
            ]))

    def run_commands(self, commands: list[str]):
        for arg_str in commands:
            try:
                self.run_command(arg_str)
            except (KeyboardInterrupt, exceptions.ShellExit):
                self.runner.exit()
                return
            except exceptions.ShellContinue:
                continue
            except exceptions.ShellUndo as undo:
                self.undo(undo.num)
                # pdf.reread(backup_path)
            except exceptions.ShellExport as export:
                self.export(export.filename)
            except exceptions.ExpectedError as e:
                print(e)
            except Exception as e:
                utils.print_unexpected_exception(e, Config.Debug)

    def run(self, commands: list[str] | None = None):
        if commands is None:
            commands = input_generator()

        self.run_commands(commands)



from pdfriend.commands.version import version
from pdfriend.commands.merge import merge
from pdfriend.commands.invert import invert
from pdfriend.commands.edit import edit
from pdfriend.commands.tinker import tinker
from pdfriend.commands.cache import cache
from pdfriend.commands.weave import weave
from pdfriend.commands.split import split
from pdfriend.commands.encrypt import encrypt
from pdfriend.commands.decrypt import decrypt
from pdfriend.commands.metadata import metadata
from pdfriend.commands.get import get



import pdfriend.classes.wrappers as wrappers
import pdfriend.classes.platforms as platforms
import pathlib

def split(
    pdf: wrappers.PDFWrapper,
    split_indices: list[int],
    outdir: str
):
    if 1 not in split_indices:
        split_indices = [1] + split_indices
    if pdf.len() not in split_indices:
        split_indices.append(pdf.len())

    outdir_path = pathlib.Path(outdir)
    platforms.ensuredir(outdir_path)
    outfile = outdir_path.joinpath(pdf.source.stem)

    ndigits = len(str(len(split_indices) - 1))

    for i, (lower, upper) in enumerate(zip(
        split_indices[:-1], split_indices[1:]
    )):
        pdf_slice = wrappers.PDFWrapper(pages = pdf.pages[(lower-1):(upper-1)])
        pdf_slice.write(f"{outfile}-{i:0{ndigits}}.pdf")



import pdfriend.classes.exceptions as exceptions
import pdfriend.classes.wrappers as wrappers
import pdfriend.classes.info as info
from pdfriend.classes.config import Config
import pathlib
import re
from typing import Self

def to_typed(input: str, type_name: str, type_converter, name: str | None = None, err_message = ""):
    try:
        return type_converter(input)
    except Exception:
        raise exceptions.ExpectedError(
            f"value \"{input}\" could not be converted to type \"{type_name}\"{err_message}"
        )

def to_file(input: str, name: str | None = None, err_message = "") -> pathlib.Path:
    file_path = pathlib.Path(input)
    if not file_path.is_file():
        raise exceptions.ExpectedError(
            f"file \"{file_path}\" was not found{err_message}"
        )

    return file_path

def to_pdf(input: str, name: str | None = None, err_message = "") -> wrappers.PDFWrapper:
    file_path = to_file(input, err_message = err_message)

    try:
        return wrappers.PDFWrapper.Read(file_path)
    except Exception as e:
        debug_message = "\nadditional info:\n{e}" if Config.Debug else ""
        raise exceptions.ExpectedError(
            f"file \"{file_path}\" could not be read as PDF{err_message}{debug_message}"
        )


def to_shell_import(input: str, name: str | None = None, err_message = "") -> list[str]:
    file = to_file(input, name = name, err_message = err_message)
    try:
        return [
            line for line in file.read_text().split("\n")
            if line != ""
        ]
    except Exception:
        raise exceptions.ExpectedError(
            f"could not read commands from file {input}{err_message}"
        )

class CmdParser:
    def __init__(self, cmd_info: str, args: list[str]):
        self.cmd_info = cmd_info
        self.args = args
        self.current_arg = 1

    def name(self) -> str:
        return self.cmd_info.primary_name

    def short(self) -> str:
        return self.cmd_info.short_name

    def arg_str(self, arg_name: str | None) -> str:
        arg_name_str = "" if arg_name is None else f" (\"{arg_name}\")"

        return f"argument {self.current_arg}{arg_name_str}"

    def loc_str(self, arg_name: str | None) -> str:
        return f"--> in {self.arg_str(arg_name)} of command \"{self.name()}\""

    @classmethod
    def FromArgs(cls,
        program_info: info.ProgramInfo,
        args: list[str],
        no_command_message: str | None = None
    ) -> Self:
        if len(args) == 0:
            raise exceptions.ExpectedError(
                no_command_message or "no command specified"
            )

        command_name = args[0]
        command_info = program_info.get_command_info(command_name)
        if command_info is None:
            raise exceptions.ExpectedError(
                f"command \"{command_name}\" does not exist"
            )

        return CmdParser(command_info, args[1:])

    @classmethod
    def FromString(cls,
        program_info: info.ProgramInfo,
        string: str,
        no_command_message: str | None = None
    ) -> Self:
        whitespace_pattern = re.compile(r"\s+")

        return cls.FromArgs(
            program_info,
            re.split(whitespace_pattern, string),
            no_command_message = no_command_message
        )

    def ensure_next_exists(self, name: str | None = None):
        if len(self.args) == 0:
            raise exceptions.ExpectedError(
                f"{self.arg_str(name)} for command \"{self.name()}\" not provided"
            )

    def advance(self, head, tail):
        self.args = tail
        self.current_arg += 1
        return head

    def split_head(self, name: str | None = None):
        self.ensure_next_exists(name = name)

        return (self.args[0], self.args[1:])

    def next_str(self, name: str | None = None):
        head, tail = self.split_head(name = name)

        return self.advance(head, tail)

    def next_str_or(self, default: str, name: str | None = None) -> str:
        try:
            return self.next_str(name)
        except Exception:
            return default

    def next_typed(self, type_name: str, type_converter, name: str | None = None):
        head, tail = self.split_head(name = name)

        return self.advance(
            to_typed(
                head, type_name, type_converter,
                name = name,
                err_message = f"\n{self.loc_str(name)}"
            ),
            tail
        )

    def next_typed_or(self, type_name: str, type_converter, default, name: str | None = None):
        try:
            return self.next_typed(type_name, type_converter, name)
        except Exception:
            return default

    def next_int(self, name: str | None = None) -> int:
        return self.next_typed("int", int, name)

    def next_int_or(self, default: int, name: str | None = None) -> int:
        try:
            return self.next_int(name)
        except Exception:
            return default

    def next_float(self, name: str | None = None) -> float:
        return self.next_typed("float", float, name = name)

    def next_float_or(self, default: float, name: str | None = None) -> float:
        try:
            return self.next_float(name = name)
        except Exception:
            return default

    def next_pdf_slice(self, pdf: wrappers.PDFWrapper, name: str | None = None) -> list[int]:
        return self.next_typed("PDF slice", lambda s: pdf.slice(s), name = name)

    def next_pdf_slice_or(self, pdf: wrappers.PDFWrapper, default: list[int], name: str | None = None) -> list[int]:
        try:
            return self.next_pdf_slice(pdf, name = name)
        except Exception:
            return default

    def next_file(self, name: str | None = None) -> pathlib.Path:
        head, tail = self.split_head(name = name)

        return self.advance(
            to_file(head, name, err_message = f"\n{self.loc_str(name)}"),
            tail
        )


    def next_file_or(self, default: pathlib.Path, name: str | None = None) -> pathlib.Path:
        try:
            return self.next_file(name = name)
        except Exception:
            return default

    def next_pdf(self, name: str | None = None) -> wrappers.PDFWrapper:
        head, tail = self.split_head(name = name)

        return self.advance(
            to_pdf(head, name, err_message = f"\n{self.loc_str(name)}"),
            tail
        )

    def next_pdf_or(self, default: wrappers.PDFWrapper, name: str | None = None) -> wrappers.PDFWrapper:
        try:
            return self.next_pdf(name = name)
        except Exception:
            return default




import pdfriend.classes.wrappers as wrappers
import pdfriend.classes.exceptions as exceptions
import pdfriend.classes.cmdparsers as cmdparsers
import pdfriend.classes.shells as shells
import pdfriend.classes.info as info
from pdfriend.classes.platforms import Platform
from pdfriend.classes.config import Config
from pypdf._utils import ImageFile
import pdfriend.utils as utils
import pathlib
from pprint import pprint

program_info = info.ProgramInfo(
    info.CommandInfo("help", "h", descr = """[command?]
    display help message. If given a command, it will only display the help message for that command.

    examples:
        help rotate
            displays the help blurb for the rotate command
        help exit
            displays the help blurb for the exit command
    """),
    info.CommandInfo("exit", "e", descr = """
    exits the edit mode
    """),
    info.CommandInfo("undo", "u", descr = """[number?]
    undo the previous [number] commands.

    examples:
        u
            undoes the previous command
        u 3
            undoes the previous 3 commands
        u all
            undoes all commands issued this session (reverts document fully)
    """),
    info.CommandInfo("export", "x", descr = """[filename?=pdfriend_tinker.txt]
    exports all the commands you ran into a text file

        examples:
            x
                exports your commands to pdfriend_edit.txt
            x out.txt
                exports your commands to out.txt
    """),
    info.CommandInfo("page", "p", descr = """[page]
    focuses on the given page

        examples:
            p 2
                focuses on page 2
    """),
    info.CommandInfo("list", "ls", descr = """
    lists objects on the current focused page (currently only works with images)

        examples:
            ls
                lists objects on the current page
    """),
    info.CommandInfo("show", "s", descr = """[name]
    opens the selected object as a separate file (currently only works with images)

        examples:
            s image.jpg
                opens the image as if it was a separate file
            s Im4.png
                same as above
    """),
    info.CommandInfo("write", "w", descr = """[name] [filename?=name]
    writes the selected object to the given file (currently only works with images)

        examples:
            w image.jpg
                writes image.jpg to ./image.jpg
            w Im7.png image.png
                writes Im7.png to ./image.png
    """),
    foreword = "pdfriend edit shell for quick changes. Commands:",
    postword = "use h [command] to learn more about a specific command"
)


class TinkerRunner(shells.ShellRunner):
    def __init__(self, pdf: wrappers.PDFWrapper, open_pdf: bool = True, open_page: bool = True):
        backup_path = pdf.backup()
        print(f"editing {pdf.source}\nbackup created in {backup_path}")

        self.pdf = pdf
        self.backup_path = backup_path
        self.open_pdf = open_pdf
        self.open_page = open_page

        self.current_page_num = None
        self.current_page_pdf = None
        self.current_page_path = Platform.NewTemp("current_page.pdf")

        if open_pdf:
            Platform.OpenFile(pdf.source)

    def current_page(self):
        return self.current_page_pdf[1]

    def raise_if_no_page(self):
        if self.current_page_pdf is None:
            raise exceptions.ExpectedError(
                "no page selected! Use page [page_number] to select a page."
            )

    def get_object(self, obj_name):
        objects = [
            obj for obj in self.current_page().images
            if obj.name == obj_name
        ]
        if len(objects) == 0:
            raise exceptions.ExpectedError(
                f"No object named {obj_name} in page {self.current_page_num}"
            )
        if len(objects) > 1:
            print(f"warning: more than one object named {obj_name} in page {self.current_page_num}")
        return objects[0]

    def write_object(self, obj, output_path: pathlib.Path | None = None):
        if output_path is None:
            output_path = Platform.NewTemp(obj.name)

        if isinstance(obj, ImageFile):
            with open(output_path, "wb") as outfile:
                outfile.write(obj.data)

        return output_path

    def parse(self, arg_str) -> list[str]:
        return utils.parse_command_string(arg_str)

    def run(self, args: list[str]):
        cmd_parser = cmdparsers.CmdParser.FromArgs(
            program_info,
            args,
            no_command_message = "No command specified! Type h or help for a list of the available commands"
        )
        short = cmd_parser.short()

        if short == "h":
            subcommand = cmd_parser.next_str_or(None)
            print(program_info.help(subcommand))

            # this is to prevent rewriting the file and appending
            # the command to the command stack
            raise exceptions.ShellContinue()
        elif short == "e":
            raise exceptions.ShellExit()
        elif short == "u":
            # arg will be converted to int, unless it's "all". Defaults to 1
            num_of_commands = cmd_parser.next_typed_or(
                "int or \"all\"", lambda s: s if s == "all" else int(s),
                1 # default value
            )

            raise exceptions.ShellUndo(num_of_commands)
        elif short == "x":
            filename = cmd_parser.next_str_or("pdfriend_edit.txt")

            raise exceptions.ShellExport(filename)
        elif short == "p":
            page_num = cmd_parser.next_int()
            self.pdf.raise_if_out_of_range(page_num)

            if self.current_page_pdf is None:
                current_page_pdf = wrappers.PDFWrapper(pages = [self.pdf[page_num]])
                current_page_pdf.write(self.current_page_path)
                self.current_page_pdf = wrappers.PDFWrapper.Read(self.current_page_path)

                if self.open_page:
                    Platform.OpenFile(self.current_page_path)
            else:
                self.current_page_pdf[1] = self.pdf[page_num]

            self.current_page_num = page_num
        elif short == "ls":
            self.raise_if_no_page()
            subcommand = cmd_parser.next_str_or(None, name = "subcommand")
            long = subcommand == "-l"

            for image in self.current_page().images:
                extra = ""
                if long:
                    extra = f"    {len(image.data) / 1000} KB"

                print(f"{image.name}{extra}")

            raise exceptions.ShellContinue()
        elif short == "s":
            self.raise_if_no_page()
            image_name = cmd_parser.next_str("name")

            image = self.get_object(image_name)
            image_path = self.write_object(image)
            Platform.OpenFile(image_path)

            raise exceptions.ShellContinue()
        elif short == "w":
            self.raise_if_no_page()
            image_name = cmd_parser.next_str(name = "name")
            filename = cmd_parser.next_str_or(image_name, name = "filename")

            image = self.get_object(image_name)
            self.write_object(image, pathlib.Path(filename))

            raise exceptions.ShellContinue()
    def reset(self):
        self.pdf.reread(self.backup_path)

    def save(self):
        if self.current_page_pdf is None:
            return
        self.pdf[self.current_page_num] = self.current_page()
        self.current_page_pdf.write()

    def exit(self):
        self.pdf.write()


def tinker(
    pdf: wrappers.PDFWrapper,
    commands: list[str] | None = None,
    open_pdf: bool = True,
    open_page: bool = True
):
    tinker_shell = shells.Shell(
        runner = TinkerRunner(pdf, open_pdf = open_pdf, open_page = open_page)
    )

    tinker_shell.run(commands)




import glob
import os
import pdfriend.classes.wrappers as wrappers
from pdfriend.classes.platforms import Platform

IMAGE_FORMATS = {".png", ".jpg"}


# expands all globs in list: can contain both patterns and single filenames,
# doesn't matter, it will eat them all
def globs_to_filenames(globs: list[str]) -> list[str]:
    result = []
    for g in globs:
        result.extend(glob.glob(g))

    return result


def images_to_pdf(filenames: list[str], outfile: str, quality: int):
    images = wrappers.ImageWrapper.FromFiles(filenames)
    images.equalize_widths()
    images.write(outfile, quality)


# unfortunately, the easiest way to do this is to write the pdf into a temporary
# file and read it again
def images_to_pdf_wrapper(filenames: list[str], quality: int) -> wrappers.PDFWrapper:
    tempfile: str = Platform.NewTemp("temp.pdf").as_posix()

    images_to_pdf(filenames, tempfile, quality)
    pdf = wrappers.PDFWrapper.Read(tempfile)
    os.remove(tempfile)

    return pdf


def merge(globs: list[str], outfile: str, quality: int):
    filenames = globs_to_filenames(globs)

    final_pdf = wrappers.PDFWrapper()
    buffer = []  # for images to be merged into pdf

    for filename in filenames:
        _, extension = os.path.splitext(filename)

        if extension in IMAGE_FORMATS:
            buffer.append(filename)
        else:  # when the block of images ends, merge them into a pdf
            if len(buffer) > 0:
                pdf = images_to_pdf_wrapper(buffer, quality)
                buffer = []

                final_pdf.merge_with(pdf)

            if extension == ".pdf":
                pdf = wrappers.PDFWrapper.Read(filename)
                final_pdf.merge_with(pdf)

            # note that if it's neither image nor a pdf, it is ignored

    if len(buffer) > 0:  # in case the filename list ends with images
        # this is for efficiency, in case all files are images: otherwise
        # backend.images_to_df_wrapper would write the pdf, read it, delete it
        # and then final_pdf.write would write it again. Instead we just write
        if final_pdf.len() == 0:
            images_to_pdf(buffer, outfile, quality)
            return

        pdf = images_to_pdf_wrapper(buffer, quality)
        final_pdf.merge_with(pdf)

    final_pdf.write(outfile)



import getpass
import pdfriend.classes.wrappers as wrappers

def encrypt(pdf: str, outfile: str):
    password = getpass.getpass("password: ")
    # if the input doesn't get overwritten, there's no need to
    # double check the password
    if pdf.source == outfile:
        confirmation = getpass.getpass("repeat the password: ")
        if password != confirmation:
            print("the passwords don't match!")
            return

    writer = pdf.to_writer()
    writer.encrypt(password)
    writer.write(pathlib.Path(outfile).with_suffix(".pdf"))




import pdfriend.classes.wrappers as wrappers
import pdfriend.classes.exceptions as exceptions

def metadata(
    pdf: wrappers.PDFWrapper,
    get: str | None = None,
    set_str: str | None = None,
    pop: str | None = None
):
    if get is not None:
        try:
            print(pdf.metadata[get])
        except KeyError:
            raise exceptions.ExpectedError(
                f"key \"{get}\" not found in the metadata.\navailable keys: {', '.join(list(pdf.metadata))}"
            )
    elif set_str is not None:
        for set_pair in set_str.split(","):
            set_pair_split = set_pair.split("=")
            if len(set_pair_split) < 2:
                raise exceptions.ExpectedError(f"no value provided for key {set_pair}")

            try:
                key, val = set_pair_split[0], set_pair_split[1]
                pdf.metadata[key] = val
                pdf.write()
            except Exception:
                raise exceptions.ExpectedError(f"metadata key \"{key}\" is invalid")
    elif pop is not None:
        keys_to_pop = pop.split(",")
        for key in keys_to_pop:
            del pdf.metadata[key]

        pdf.write(infile)
    else:
        print("\n".join([f"{key}: {val}" for key, val in pdf.metadata.items()]))


